package com.charego.rush

import java.lang.RuntimeException

fun main() {
    val startPosition = loadBoard("/boards/board3.txt")
    println("Input:")
    printBoard(startPosition.board)

    val finalPosition = solve(startPosition, maxDepth = 50)
    println("Solution (n=${finalPosition.previousBoards.size}):")
    //for (board in finalPosition.previousBoards) {
    //    printBoard(board)
    //}
    printBoard(finalPosition.board)
}

/* Breadth-first search */
fun solve(startPosition: Position, maxDepth: Int): Position {
    val visitedBoards = mutableSetOf(startPosition.id)
    val queue = ArrayDeque(setOf(startPosition))

    while (queue.isNotEmpty()) {
        val position = queue.removeFirst()
        if (position.isSolution()) {
            return position
        }
        if (position.previousBoards.size + 1 >= maxDepth) {
            continue
        }
        for (child in position.findAvailableMoves()) {
            if (!visitedBoards.contains(child.id)) {
                queue.addLast(child)
                visitedBoards.add(child.id)
            }
        }
    }

    // If we exit the loop without finding a solution, throw an exception
    throw RuntimeException("No solution could be found up to $maxDepth moves!")
}

fun loadBoard(resourceName: String): Position {
    val boardString = {}::class.java.getResource(resourceName).readText().trim()
    val board = boardString.split(Regex("\\s+")).map { x -> x.toInt() }.toIntArray()
    return Position(board, emptyList())
}

fun printBoard(board: IntArray) {
    fun printRow(startIndex: Int, endIndex: Int) {
        val row = board.slice(startIndex..endIndex)
        val fmt = "    %2d %2d %2d %2d %2d %2d"
        val res = fmt.format(*row.toTypedArray())
        println(res)
    }
    printRow(0, 5)
    printRow(6, 11)
    printRow(12, 17)
    printRow(18, 23)
    printRow(24, 29)
    printRow(30, 35)
    println()
}
