package com.charego.rush

/**
 * Models a Rush Hour game board as an array of integers with indices:
 *
 * ```
 *  0  1  2  3  4  5
 *  6  7  8  9 10 11
 * 12 13 14 15 16 17
 * 18 19 20 21 22 23
 * 24 25 26 27 28 29
 * 30 31 32 33 34 35
 * ```
 */
class Position(val board: IntArray, val previousBoards: List<IntArray>) {
    val id = board.contentHashCode()

    fun isSolution(): Boolean {
        return board[16] != 0 && board[16] == board[17]
    }

    fun findAvailableMoves(): List<Position> {
        val availableMoves = mutableListOf<Position>()
        for ((spaceIndex, value) in board.withIndex()) {
            if (value == 0) {
                val rowIndex = spaceIndex / 6
                val colIndex = spaceIndex % 6
                //println("$spaceIndex: (${rowIndex}, ${colIndex})")
                //println("ROW: ${board.sliceArray((rowIndex * 6) until ((rowIndex + 1) * 6)).contentToString()}")
                //println("COL: ${board.slice(colIndex .. (30 + colIndex) step 6).toIntArray().contentToString()}")
                if (colIndex > 1) {
                    val lookLeft = board.sliceArray((rowIndex * 6) until spaceIndex).reversedArray()
                    //println("LOOKING LEFT: ${lookLeft.contentToString()}")
                    val carData = findCar(lookLeft)
                    if (carData.size > 1) {
                        //println("CAR: $carData")
                        val newBoard = board.copyOf()
                        val startIndex = spaceIndex - carData.size - carData.distance
                        for (i in startIndex .. spaceIndex) {
                            newBoard[i] = 0
                        }
                        for (i in (spaceIndex - carData.size + 1) .. spaceIndex) {
                            newBoard[i] = carData.number
                        }
                        //println("UPDATED ROW: ${newBoard.sliceArray((rowIndex * 6) until ((rowIndex + 1) * 6)).contentToString()}")
                        availableMoves.add(moveTo(newBoard))
                    }
                }
                if (colIndex < 4) {
                    val lookRight = board.sliceArray((spaceIndex + 1) until ((rowIndex + 1) * 6))
                    //println("LOOKING RIGHT: ${lookRight.contentToString()}")
                    val carData = findCar(lookRight)
                    if (carData.size > 1) {
                        //println("CAR: $carData")
                        val newBoard = board.copyOf()
                        val endIndex = spaceIndex + carData.size + carData.distance
                        for (i in spaceIndex .. endIndex) {
                            newBoard[i] = 0
                        }
                        for (i in spaceIndex until (spaceIndex + carData.size)) {
                            newBoard[i] = carData.number
                        }
                        //println("UPDATED ROW: ${newBoard.sliceArray((rowIndex * 6) until ((rowIndex + 1) * 6)).contentToString()}")
                        availableMoves.add(moveTo(newBoard))
                    }
                }
                if (rowIndex > 1) {
                    val lookUp = board.slice(colIndex .. (spaceIndex - 6) step 6).toIntArray().reversedArray()
                    //println("LOOKING UP: ${lookUp.contentToString()}")
                    val carData = findCar(lookUp)
                    if (carData.size > 1) {
                        //println("CAR: $carData}")
                        val newBoard = board.copyOf()
                        for (i in (spaceIndex - (carData.size + carData.distance) * 6) .. (spaceIndex - carData.size * 6) step 6) {
                            newBoard[i] = 0
                        }
                        for (i in (spaceIndex - (carData.size - 1) * 6) .. spaceIndex step 6) {
                            newBoard[i] = carData.number
                        }
                        //println("UPDATED COL: ${newBoard.slice(colIndex .. (30 + colIndex) step 6).toIntArray().contentToString()}")
                        availableMoves.add(moveTo(newBoard))
                    }
                }
                if (rowIndex < 4) {
                    val lookDown = board.slice((spaceIndex + 6) .. (colIndex + 30) step 6).toIntArray()
                    //println("LOOKING DOWN: ${lookDown.contentToString()}")
                    val carData = findCar(lookDown)
                    if (carData.size > 1) {
                        //println("CAR: $carData")
                        val newBoard = board.copyOf()
                        for (i in spaceIndex .. (spaceIndex + (carData.size - 1) * 6) step 6) {
                            newBoard[i] = carData.number
                        }
                        for (i in (spaceIndex + carData.size * 6) .. (spaceIndex + ((carData.size + carData.distance) * 6)) step 6) {
                            newBoard[i] = 0
                        }
                        //println("UPDATED COL: ${newBoard.slice(colIndex .. (30 + colIndex) step 6).toIntArray().contentToString()}")
                        availableMoves.add(moveTo(newBoard))
                    }
                }
            }
        }
        //println("Found ${availableMoves.size} moves...")
        return availableMoves
    }

    private fun moveTo(newBoard: IntArray): Position {
        val previousBoardsCopy = ArrayList(previousBoards)
        previousBoardsCopy.add(board)
        return Position(newBoard, previousBoardsCopy)
    }

    data class CarData(val size: Int, val number: Int, val distance: Int)

    private fun findCar(view: IntArray): CarData {
        var carSize = 0
        var carNumber = -1
        var carDistance = 0
        for (number in view) {
            if (carNumber == -1) {
                if (number == 0) {
                    carDistance += 1
                } else {
                    carNumber = number
                    carSize = 1
                }
            } else if (carNumber == number) {
                carSize += 1
            } else {
                break
            }
        }
        return CarData(carSize, carNumber, carDistance)
    }
}
